/*
** benchmark.d
** Copyright (C) 2003 Helmut Leitner
** Donated to Digital Mars.
** Released under the GNU GPL, see LICENSE.TXT
** and http://www.gnu.org/copyleft/gpl.html
*/

module venus.benchmark;
import venus.all;

double BenchLimit=0.001;
double BenchLimitAboveStat=3.0;
int BenchDim=100;

double DelegateRetSec(void delegate () dg)
{
    ulong t1,t2;
    double sec;

    TimerGetCount(&t1);
    dg();
    TimerGetCount(&t2);
    sec=TimerCountRetSec(t2-t1);
    return sec;
}

double DelegateCountRetSec(void delegate () dg,int count)
{
    ulong t1,t2,t3,d1,d2;
    double sec;
    int i;

do_restart:
    TimerGetCount(&t1);
    for(i=0; i<count; i++) {
        dg();
    }
    TimerGetCount(&t2);
    for(i=0; i<count; i++) {
        dg();
        dg();
    }
    TimerGetCount(&t3);
    d1=t2-t1;
    d2=t3-t2;
    if(d2<=d1) {
        goto do_restart;
    }
    sec=TimerCountRetSec(d2-d1);
    return sec;
}

void DoubleArPrintFmt(double [] dtab,char *fmt)
{
    for(int i; i<dtab.length; i++) {
        printf(fmt,i,dtab[i]);
    }
}

double DelegateBenchRetSec(void delegate () dg)
{
    double sec;
    int i,mid=BenchDim/2,range=mid/2;
    double sum;
    int sumn;
    int loop=0;
    ulong t1,t2;
    int sectabdim=BenchDim;

    int count=1;

    sec=DelegateRetSec(dg);
    if(sec>BenchLimitAboveStat) {
       return sec;
    } else if(sec>BenchLimit) {
       sectabdim=3*BenchLimitAboveStat/sec;
       if(sectabdim<8) {
          sectabdim=8;
       }
    } else {
       do {
          sec=DelegateCountRetSec(dg,count);
          if(sec>=BenchLimit && sec<100.0) {
             break;
          }
          count*=2;
       } while(sec<BenchLimit);
    }

    double sectab[]=new double [sectabdim];

//    printf("count=%d sec=%lf\n",count,sec);
    for(i=0; i<sectab.length; i++) {
        sectab[i]=DelegateCountRetSec(dg,count)/(double)count;
    }

    // sorting is a bit complicated
//    DoubleArPrintFmt(sectab,"sectab[%d] %.10lf\n");
    for(i=0; i<sectab.length; i++) {
        sectab[i]*=1e12; // avoid .sort Error with small numbers
    }
    sectab.sort;
    for(i=0; i<sectab.length; i++) {
        sectab[i]/=1e12;
    }
//    DoubleArPrintFmt(sectab,"sectab[%d] %.10lf\n");
    sum=0.0;
    for(i=mid-range; i<mid+range; i++) {
        sum += sectab[i]; sumn++;
    }

do_return:
    sec=sum/sumn;
    return sec;
}

double DelegateBenchPrint(void delegate () dg,char name[])
{
    double sec=DelegateRetSec(dg);
    printf("%.*s %.3lf usec\n",name,sec*1e6);
    return sec;
}

double FpRetSec(void (*fp)())
{
    void_dg_void dg=FpRetDelegate(fp);
    return DelegateRetSec(dg);
}

double FpBenchPrint(void (*fp)(),char name[])
{
    void_dg_void dg=FpRetDelegate(fp);
    return DelegateBenchPrint(dg,name);
}

